/*
  For more information, please see: http://software.sci.utah.edu

  The MIT License

  Copyright (c) 2004 Scientific Computing and Imaging Institute,
  University of Utah.

  License for the specific language governing rights and limitations under
  Permission is hereby granted, free of charge, to any person obtaining a
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
*/



/*
 *  Object.h: Base class for all PIDL distributed objects
 *
 *  Written by:
 *   Steven G. Parker
 *   Department of Computer Science
 *   University of Utah
 *   July 1999
 *
 *  Copyright (C) 1999 SCI Group
 */

#ifndef CCA_PIDL_Object_h
#define CCA_PIDL_Object_h


#include <Core/CCA/PIDL/URL.h>
#include <Core/CCA/PIDL/ReferenceMgr.h>
#include <Core/CCA/PIDL/MxNArrayRep.h>
#include <Core/CCA/PIDL/MxNScheduler.h>
#include <Core/CCA/PIDL/HandlerStorage.h>
//#include <Core/CCA/PIDL/HandlerGateKeeper.h>
#include <Core/CCA/PIDL/EpChannel.h>
#include <Core/CCA/SmartPointer.h>


namespace SCIRun {

class MutexPool;
class TypeInfo;
class ServerContext;
class Reference;
class URL;

/**************************************

CLASS
   Object

KEYWORDS
   Object, PIDL

DESCRIPTION
   The base class for all PIDL based distributed objects.  It provides
   a single public method - getURL, which returns a URL for that object.
   The _getReference method is also in the public section, but should
   be considered private to the implementation and should not be used
   outside of PIDL or sidl generated stubs.

****************************************/

class Object {
public:
  typedef CCALib::SmartPointer<Object> pointer;

  //////////
  // Destructor
  virtual ~Object();

  //////////
  // Return a URL that can be used to contact this object.
  URL getURL() const;

  //////////
  // Internal method to get a reference (startpoint and
  // vtable base) to this object.
  virtual void _getReferenceCopy(ReferenceMgr** ) const;

  //////////
  // Internal method to increment the reference count for
  // this object.
  void _addReference();

  //////////
  // Internal method to decrement the reference count for
  // this object.
  void _deleteReference();

  //////////
  // External method to increment the reference count for
  // this object.
  void addReference();

  //////////
  // External method to decrement the reference count for
  // this object.
  void deleteReference();

  //////////
  // Used to set the distribution of a particular array
  // when this object is the callee
  void setCalleeDistribution(std::string distname,
                             MxNArrayRep* arrrep);

  //////////
  // The context of the server object.  If this is null,
  // then the object is a proxy and there is no server.
  ServerContext* d_serverContext;

  //////////
  // Method used to create an array distribution scheduler
  void createScheduler();

  /////////
  // Methods used only on proxies and generated by SIDL
  // (no implementation for server)
  void createSubset(int, int);
  void setRankAndSize(int, int);
  void resetRankAndSize();

  ////////
  // Dummy method (needed by PIDL/proxy_sample
  virtual void getException();

  //DEBUG:: remove?
  int getReferenceCount();
protected:
  //////////
  // Constructor.  Initializes d_serverContext to null,
  // which means that it is a proxy object.  Calls to
  // intializeServer from derived class constructors will
  // set up d_serverContext appropriately.
  Object();

  //////////
  // Set up d_serverContext with the given type information
  // and the void pointer to be passed back to the generated
  // stub code.  This will get called by each constructor
  // in the entire inheritance tree.  The last one to call
  // will be the most derived class, which is the only
  // one that we care about.
  void initializeServer(const TypeInfo* typeinfo, void* ptr, EpChannel* epc);

private:

  friend class Warehouse;

  //////////
  // The reference count for this object.
  int ref_cnt;

  //////////
  // The index of the mutex in the mutex pool.
  // (not currently used)
  int mutex_index;

  //////////
  // Activate the server.  This registers the object with the
  // object wharehouse, and creates the globus endpoint.
  // This is not done until something requests a reference
  // to the object though getURL() or _getReference().
  void activateObject() const;

  //////////
  // Return the singleton Mutex pool shared by all instances
  // of Object.
  static SCIRun::MutexPool* getMutexPool();

  static Mutex sm;

  //////////
  // Private copy constructor to make copying impossible.
  Object(const Object&);

  //////////
  // Private assignment operator to make assignment impossible.
  Object& operator=(const Object&);

  int objid;
  static int objcnt;


  //FOR TEST ONLY, will remove later
  bool firstTime;
};

} // End namespace SCIRun

#endif
