/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2004 Scientific Computing and Imaging Institute,
   University of Utah.

   License for the specific language governing rights and limitations under
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/


/*
 *  BabelComponentInstance.cc:
 *
 *  Written by:
 *   Keming Zhang
 *   Department of Computer Science
 *   University of Utah
 *   July 2002
 *
 */

#include <Framework/SCIRunFramework.h>
#include <Framework/Babel/BabelPortInstance.h>
#include <Framework/Babel/BabelComponentInstance.h>
#include <Framework/CCA/CCAPortInstance.h>
#include <Framework/CCA/CCAException.h>
#include <Core/CCA/spec/cca_sidl.h>
#include <Framework/TypeMap.h>

#include <iostream>

#ifndef DEBUG
#  define DEBUG 0
#endif

namespace SCIRun {

BabelCCAGoPort::BabelCCAGoPort(const UCXX ::gov::cca::ports::GoPort& port)
{
  this->port = port;
}

int BabelCCAGoPort::go()
{
  return port.go();
}

BabelCCAUIPort::BabelCCAUIPort(const UCXX ::gov::cca::ports::UIPort& port)
{
  this->port = port;
}

int BabelCCAUIPort::ui()
{
  return port.ui();
}


BabelComponentInstance::BabelComponentInstance(SCIRunFramework* framework,
                                               const std::string& instanceName,
                                               const std::string& typeName,
                                               const UCXX ::gov::cca::TypeMap& properties,
                                               const UCXX ::gov::cca::Component& component,
                                               const UCXX ::framework::Services& svc)
  : ComponentInstance(framework, instanceName, typeName, sci::cca::TypeMap::pointer(0))
{
  // Babel component properties are ignored for now.
  this->component = component;
  this->svc = svc;

  BabelPortInstance *go = dynamic_cast<BabelPortInstance*>(getPortInstance("go"));
  if (go != 0) {
    // locate a GoPort provided by the Babel component instance
    PortInstanceMap *pports = (PortInstanceMap*) (this->svc.getData());
    UCXX ::gov::cca::ports::GoPort g = UCXX ::sidl::babel_cast<UCXX ::gov::cca::ports::GoPort>(go->getPort());
    sci::cca::ports::GoPort::pointer goPort(new BabelCCAGoPort(g));
    CCAPortInstance *piGo = new CCAPortInstance("go",
						"sci.cca.ports.GoPort",
						sci::cca::TypeMap::pointer(0),
						goPort,
						CCAPortInstance::Provides);

    // Override the original gov.cca.GoPort in the pports map
    // (the original port is stored in a BabelCCAGoPort) instantiated above.
    // This is needed to connect GoPorts with CCA UI components such as the GUIBuilder.
    (*pports)["go"] = (PortInstance*) piGo;
  }

  // locate a UIPort provided by the Babel component instance
  BabelPortInstance *ui = dynamic_cast<BabelPortInstance*>(getPortInstance("ui"));
  if (ui != 0) {
    PortInstanceMap *pports= (PortInstanceMap*) (this->svc.getData());

    // If there is a UIPort, create a corresponding CCAPortInstance
    UCXX ::gov::cca::ports::UIPort u = UCXX ::sidl::babel_cast<UCXX ::gov::cca::ports::UIPort>(ui->getPort());
    sci::cca::ports::UIPort::pointer uiPort(new BabelCCAUIPort(u));
    CCAPortInstance *piUI = new CCAPortInstance("ui",
						"sci.cca.ports.UIPort",
						sci::cca::TypeMap::pointer(0),
						uiPort,
						CCAPortInstance::Provides);

    // Override the original gov.cca.UIPort in the pports map
    // (the original port is stored in a BabelCCAUIPort) instantiated above.
    // This is needed to connect UIPorts with CCA UI components such as the GUIBuilder.
    (*pports)["ui"] = (PortInstance*) piUI;
  }
}

BabelComponentInstance::~BabelComponentInstance()
{
#if DEBUG
  std::cerr << "BabelComponentInstance destroyed..." << std::endl;
#endif
}

PortInstance*
BabelComponentInstance::getPortInstance(const std::string& portname)
{
  PortInstanceMap *pports = (PortInstanceMap*) svc.getData();

  if (pports != 0) {
    PortInstanceMap::iterator iter = pports->find(portname);
    if (iter == pports->end()) {
      return 0;
    } else {
      return iter->second;
    }
  } else {
    std::cerr << "Warning: NULL pports!" << std::endl;
    return 0;
  }
}

UCXX ::gov::cca::Port
BabelComponentInstance::getPort(const std::string& name)
{
  return svc.getPort(name);
}

UCXX ::gov::cca::Port
BabelComponentInstance::getPortNonblocking(const std::string& name)
{
  return svc.getPortNonblocking(name);
}

void
BabelComponentInstance::releasePort(const std::string& name)
{
  return svc.releasePort(name);
}

UCXX ::gov::cca::TypeMap
BabelComponentInstance::createTypeMap()
{
  return svc.createTypeMap();
}

void
BabelComponentInstance::registerUsesPort(const std::string& portName,
                                         const std::string& portType,
                                         const UCXX ::gov::cca::TypeMap& properties)
{
  return svc.registerUsesPort(portName, portType, properties);
}

void BabelComponentInstance::unregisterUsesPort(const std::string& name)
{
  return svc.unregisterUsesPort(name);
}

void
BabelComponentInstance::addProvidesPort(const UCXX ::gov::cca::Port& port,
                                        const std::string& portName,
                                        const std::string& portType,
                                        const UCXX ::gov::cca::TypeMap& properties)
{
  return svc.addProvidesPort(port, portName, portType, properties);
}

void BabelComponentInstance::removeProvidesPort(const std::string& name)
{
  svc.removeProvidesPort(name);
  return;
}

// gov::cca::TypeMap BabelComponentInstance::getPortProperties(const std::string& portName)
// {
//   return svc.getPortProperties(portName);
// }
// these do nothing at the moment - implement after compiler changeover
sci::cca::TypeMap::pointer
BabelComponentInstance::getPortProperties(const std::string& portName)
{
  return sci::cca::TypeMap::pointer(new TypeMap);
}

UCXX ::gov::cca::ComponentID
BabelComponentInstance::getComponentID()
{
  return svc.getComponentID();
}

PortInstanceIterator*
BabelComponentInstance::getPorts()
{
  return new Iterator(this);
}

BabelComponentInstance::Iterator::Iterator(BabelComponentInstance* comp)
{
  ports = (PortInstanceMap*) (comp->svc.getData());
  iter = ports->begin();
}

BabelComponentInstance::Iterator::~Iterator()
{
}

PortInstance* BabelComponentInstance::Iterator::get()
{
  return iter->second;
}

bool BabelComponentInstance::Iterator::done()
{
  return iter == ports->end();
}

void BabelComponentInstance::Iterator::next()
{
  ++iter;
}

} // end namespace SCIRun
