/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2004 Scientific Computing and Imaging Institute,
   University of Utah.

   License for the specific language governing rights and limitations under
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  NrrdReader.cc: Read in a Nrrd
 *
 *  Written by:
 *   David Weinstein
 *   School of Computing
 *   University of Utah
 *   February 2001
 *
 *  Copyright (C) 2001 SCI Group
 */

#include <Dataflow/Network/Module.h>
#include <Core/Malloc/Allocator.h>
#include <Core/GuiInterface/GuiVar.h>
#include <Dataflow/Network/Ports/NrrdPort.h>
#include <Core/Util/sci_system.h>
#include <Core/Containers/StringUtil.h>
#include <sys/stat.h>
#include <sstream>
#ifdef _WIN32
#include <process.h> // for getpid
#endif

namespace SCITeem {

using namespace SCIRun;

class NrrdReader : public Module {
public:
  NrrdReader(SCIRun::GuiContext* ctx);
  virtual ~NrrdReader();
  virtual void execute();

private:
  bool read_nrrd();
  bool read_file(string filename);
  bool write_tmpfile(string filename, string *tmpfilename, string conv_command);

  GuiFilename     filename_;

  NrrdDataHandle  read_handle_;

  string          old_filename_;
  time_t          old_filemodification_;
  int             cached_label_generation_;
  char *          cached_label_;
};

} // end namespace SCITeem

using namespace SCITeem;

DECLARE_MAKER(NrrdReader)

NrrdReader::NrrdReader(SCIRun::GuiContext* ctx) : 
  Module("NrrdReader", ctx, Filter, "DataIO", "Teem"),
  filename_(get_ctx()->subVar("filename"), ""),
  read_handle_(0),
  old_filemodification_(0),
  cached_label_generation_(0),
  cached_label_(0)
{
}

NrrdReader::~NrrdReader()
{
  if (cached_label_) { delete [] cached_label_; cached_label_ = 0; }
}


// Return true if handle_ was changed, otherwise return false.  This
// return value does not necessarily signal an error!
bool
NrrdReader::read_nrrd() 
{
  filename_.reset();
  string fn(filename_.get());
  if (fn == "") { 
    error("Please specify nrrd filename");
    return false; 
  }

  // Read the status of this file so we can compare modification timestamps.
  struct stat buf;
  if (stat(fn.c_str(), &buf) == - 1) {
    error(string("FieldReader error - file not found: '")+fn+"'");
    return false;
  }

  // If we haven't read yet, or if it's a new filename, 
  //  or if the datestamp has changed -- then read...
#ifdef __sgi
  time_t new_filemodification = buf.st_mtim.tv_sec;
#else
  time_t new_filemodification = buf.st_mtime;
#endif
  if(!read_handle_.get_rep() || 
     fn != old_filename_ || 
     new_filemodification != old_filemodification_)
  {
    old_filemodification_ = new_filemodification;
    old_filename_=fn;
    read_handle_ = 0;

    int len = fn.size();
    // Filename as string
    const string filename(fn);
    const string ext(".nd");
    const string vff_ext(".vff");
    const string vff_conv_command("vff2nrrd %f %t");
    const string pic_ext(".pic");
    const string pic_ext2(".pict");
    const string pic_conv_command("PictToNrrd %f %t");
    const string vol_ext(".vol");
    const string vol_conv_command("GeoProbeToNhdr %f %t");
    const string vista_ext(".v");
    const string vista_conv_command("VistaToNrrd %f %t");


    // check that the last 3 chars are .nd for us to pio
    if (fn.substr(len - ext.size(), ext.size()) == ext) {
      Piostream *stream = auto_istream(fn, this);
      if (!stream)
      {
	error("Error reading file '" + fn + "'.");
	return true;
      }

      // Read the file
      Pio(*stream, read_handle_);
      if (!read_handle_.get_rep() || stream->error())
      {
	error("Error reading data from file '" + fn +"'.");
	delete stream;
	return true;
      }
      delete stream;
    } else { // assume it is just a nrrd
      if (fn.substr(len - vff_ext.size(), vff_ext.size()) == vff_ext){
	string tmpfilename;
	write_tmpfile(filename, &tmpfilename, vff_conv_command);
	return read_file(tmpfilename);	
	
      } else if ((fn.substr(len - pic_ext.size(), pic_ext.size()) == pic_ext) ||
		  fn.substr(len - pic_ext2.size(), pic_ext2.size()) == pic_ext2){
	string tmpfilename;
	write_tmpfile(filename, &tmpfilename, pic_conv_command);
	return read_file(tmpfilename);
	
      } else if (fn.substr(len - vol_ext.size(), vol_ext.size()) == vol_ext){
	string tmpfilename;
	write_tmpfile(filename, &tmpfilename, vol_conv_command);
	return read_file(tmpfilename);
	
      } else if (fn.substr(len - vista_ext.size(), vista_ext.size()) == vista_ext){
	string tmpfilename;
	write_tmpfile(filename, &tmpfilename, vista_conv_command);
	return read_file(tmpfilename);
	
      } else {
	return read_file(fn);
	
      }
    }
    return true;
  }
    return false;
}


bool
NrrdReader::read_file(string fn)
{

  NrrdData *n = scinew NrrdData;
  if (nrrdLoad(n->nrrd_, airStrdup(fn.c_str()), 0)) {
    char *err = biffGetDone(NRRD);
    error("Read error on '" + fn + "': " + err);
    free(err);
    return true;
  }
  read_handle_ = n;
  for (unsigned int i = 0; i < read_handle_->nrrd_->dim; i++) {
    if (!(airExists(read_handle_->nrrd_->axis[i].min) && 
	  airExists(read_handle_->nrrd_->axis[i].max)))
      nrrdAxisInfoMinMaxSet(read_handle_->nrrd_, i, nrrdCenterNode);
  }
  return false;
}


bool
NrrdReader::write_tmpfile(string filename, string* tmpfilename,
                          string conv_command)
{
  string::size_type loc = filename.find_last_of("/");
  const string basefilename =
    (loc==string::npos)?filename:filename.substr(loc+1);
	
  // Base filename with first extension removed.
  loc = basefilename.find_last_of(".");
  const string basenoext = basefilename.substr(0, loc);

  // Filename with first extension removed.
  loc = filename.find_last_of(".");
  const string noext = filename.substr(0, loc);

  // Temporary filename.
  //tmpfilename = "/tmp/" + basenoext + "-" +
  //  to_string((unsigned int)(getpid())) + ".nhdr";
  tmpfilename->assign("/tmp/" + basenoext + "-" +
		      to_string((unsigned int)(getpid())) + ".nhdr");
  
  ASSERT(sci_getenv("SCIRUN_OBJDIR"));
  string command =
    string(sci_getenv("SCIRUN_OBJDIR")) + "/StandAlone/convert/" +
    conv_command;
  while ((loc = command.find("%f")) != string::npos){
    command.replace(loc, 2, "'"+filename+"'");
  }
  while ((loc = command.find("%t")) != string::npos){
    command.replace(loc, 2, "'"+*tmpfilename+"'");
  }
  const int status = sci_system(command.c_str());
  if (status)
  {
    remark("'" + command + "' failed.  Read may not work.");
  }
  return true;
}


void
NrrdReader::execute()
{
  update_state(NeedData);

  read_nrrd();

  if (!read_handle_.get_rep()) { 
    error("Please load a nrrd.");
    return; 
  }

  // Send the data downstream.
  NrrdOPort *outport = (NrrdOPort *)get_oport("Output Data");
  outport->send_and_dereference(read_handle_, true);

  update_state(Completed);
}
