/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2004 Scientific Computing and Imaging Institute,
   University of Utah.

   License for the specific language governing rights and limitations under
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/


/*
 *  CurveFieldToText.cc
 *
 *  Written by:
 *   David Weinstein
 *   Department of Computer Science
 *   University of Utah
 *   April 2003
 *
 *  Copyright (C) 2003 SCI Group
 */

// This program will read in a SCIRun CurveField, and will save
// out the CurveMesh into two files: a .pts file and a .edge file.
// The .pts file will specify the x/y/z coordinates of each 
// point, one per line, enedgees separated by white space; the file will
// also have a one line header, specifying number of points, unless the
// user specifies the -noPtsCount command-line argument.
// The .edge file will specify the i/j indices for each edge, 
// also one per line, again with a one line header (unless a 
// -noElementsCount flag is used).  The edge indicies will be zero-based, 
// unless the user specifies -oneBasedIndexing.

#include <Core/Basis/CrvLinearLgn.h>
#include <Core/Datatypes/CurveMesh.h>
#include <Core/Datatypes/GenericField.h>
#include <Core/Containers/HashTable.h>
#include <Core/Init/init.h>
#include <Core/Persistent/Pstreams.h>

#include <StandAlone/convert/FileUtils.h>

#include <iostream>
#include <fstream>
#include <stdlib.h>

using std::cerr;
using std::ifstream;
using std::endl;

using namespace SCIRun;

bool ptsCountHeader;
int baseIndex;
bool elementsCountHeader;

void setDefaults() {
  ptsCountHeader=true;
  baseIndex=0;
  elementsCountHeader=true;
}

int parseArgs(int argc, char *argv[]) {
  int currArg = 4;
  while (currArg < argc) {
    if (!strcmp(argv[currArg],"-noPtsCount")) {
      ptsCountHeader=false;
      currArg++;
    } else if (!strcmp(argv[currArg], "-noElementsCount")) {
      elementsCountHeader=false;
      currArg++;
    } else if (!strcmp(argv[currArg], "-oneBasedIndexing")) {
      baseIndex=1;
      currArg++;
    } else {
      cerr << "Error - unrecognized argument: "<<argv[currArg]<<"\n";
      return 0;
    }
  }
  return 1;
}

void printUsageInfo(char *progName) {
  cerr << "\n Usage: "<<progName<<" CurveField pts edges [-noPtsCount] [-noElementsCount] [-oneBasedIndexing]\n\n";
  cerr << "\t This program will read in a SCIRun CurveField, and will save \n";
  cerr << "\t out the CurveMesh into two files: a .pts file and a .edge \n";
  cerr << "\t file.  The .pts file will specify the x/y/z coordinates of \n";
  cerr << "\t each point, one per line, enedgees separated by white space; \n";
  cerr << "\t the file will also have a one line header, specifying number \n";
  cerr << "\t of points, unless the user specifies the -noPtsCount \n";
  cerr << "\t command-line argument.  The .edge file will specify the i/j \n";
  cerr << "\t indices for each edge, also one per line, again with a one \n";
  cerr << "\t line header (unless a -noElementsCount flag is used).  The edge \n";
  cerr << "\t indicies will be zero-based, unless the user specifies \n";
  cerr << "\t -oneBasedIndexing.\n\n";
}

int
main(int argc, char **argv) {
  if (argc < 4 || argc > 7) {
    printUsageInfo(argv[0]);
    return 2;
  }
  SCIRunInit();
  setDefaults();

  char *fieldName = argv[1];
  char *ptsName = argv[2];
  char *edgesName = argv[3];
  if (!parseArgs(argc, argv)) {
    printUsageInfo(argv[0]);
    return 2;
  }

  FieldHandle handle;
  Piostream* stream=auto_istream(fieldName);
  if (!stream) {
    cerr << "Couldn't open file "<<fieldName<<".  Exiting...\n";
    return 2;
  }
  Pio(*stream, handle);
  if (!handle.get_rep()) {
    cerr << "Error reading surface from file "<<fieldName<<".  Exiting...\n";
    return 2;
  }
  if (handle->get_type_description(Field::MESH_TD_E)->get_name().find("CurveMesh") != 
      string::npos) 
  {
    cerr << "Error -- input field wasn't a CurveField (type_name="
	 << handle->get_type_description(Field::MESH_TD_E)->get_name() << std::endl;
    return 2;
  }
  typedef CurveMesh<CrvLinearLgn<Point> > CMesh;
  MeshHandle mH = handle->mesh();
  CMesh *cm = dynamic_cast<CMesh *>(mH.get_rep());
  CMesh::Node::iterator niter; 
  CMesh::Node::iterator niter_end; 
  CMesh::Node::size_type nsize; 
  cm->begin(niter);
  cm->end(niter_end);
  cm->size(nsize);
  FILE *fPts = fopen(ptsName, "wt");
  if (!fPts) {
    cerr << "Error opening output file "<<ptsName<<"\n";
    return 2;
  }
  if (ptsCountHeader) fprintf(fPts, "%d\n", (unsigned)(nsize));
  cerr << "Number of points = "<< nsize <<"\n";
  while(niter != niter_end) {
    Point p;
    cm->get_center(p, *niter);
    fprintf(fPts, "%lf %lf %lf\n", p.x(), p.y(), p.z());
    ++niter;
  }
  fclose(fPts);

  CMesh::Edge::size_type esize; 
  CMesh::Edge::iterator eiter; 
  CMesh::Edge::iterator eiter_end; 
  CMesh::Node::array_type edge_nodes(2);
  cm->size(esize);
  cm->begin(eiter);
  cm->end(eiter_end);
  FILE *fEdges = fopen(edgesName, "wt");
  if (!fEdges) {
    cerr << "Error opening output file "<<edgesName<<"\n";
    return 2;
  }
  if (elementsCountHeader) fprintf(fEdges, "%d\n", (unsigned)(esize));
  cerr << "Number of edges = "<< esize <<"\n";
  while(eiter != eiter_end) {
    cm->get_nodes(edge_nodes, *eiter);
    fprintf(fEdges, "%d %d\n", 
	    (int)edge_nodes[0]+baseIndex,
	    (int)edge_nodes[1]+baseIndex);
    ++eiter;
  }
  fclose(fEdges);

  return 0;  
}    
