/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2004 Scientific Computing and Imaging Institute,
   University of Utah.

   License for the specific language governing rights and limitations under
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/


/*
 *  TextToPointCloudField.cc
 *
 *  Written by:
 *   Michael Callahan
 *   Department of Computer Science
 *   University of Utah
 *   December 2003
 *
 *  Copyright (C) 2003 SCI Group
 */

// This program will read in a .pts file specifying the
// the x/y/z/s components for each point separated by
// white space.  The input file should NOT contain any
// header.  The 's' component of each line is the string
// associated with that point. The SCIRun output file is
// written in ASCII unless -binOutput is specified.

#include <Core/Datatypes/GenericField.h>
#include <Core/Basis/Constant.h>
#include <Core/Datatypes/PointCloudMesh.h>
#include <Core/Persistent/Pstreams.h>
#include <Core/Containers/HashTable.h>
#include <StandAlone/convert/FileUtils.h>
#include <Core/Init/init.h>
#include <iostream>
#include <fstream>
#include <stdlib.h>

#ifdef _WIN32
#define SHARE __declspec(dllimport)
#else
#define SHARE
#endif

namespace SCIRun {
extern SHARE FieldHandle
TextPointCloudString_reader(ProgressReporter *pr, const char *file);
}

using std::cerr;
using std::ifstream;
using std::endl;

using namespace SCIRun;


bool binOutput;
bool debugOn;

void setDefaults() {
  binOutput=false;
  debugOn=false;
}

int parseArgs(int argc, char *argv[]) {
  int currArg = 3;
  while (currArg < argc) {
    if (!strcmp(argv[currArg], "-binOutput")) {
      binOutput=true;
      currArg++;
    } else if (!strcmp(argv[currArg], "-debug")) {
      debugOn=true;
      currArg++;
    } else {
      cerr << "Error - unrecognized argument: "<<argv[currArg]<<"\n";
      return 0;
    }
  }
  return 1;
}

void printUsageInfo(char *progName) {
  cerr << "\n Usage: "<<progName<<" pts PointCloudMesh [-binOutput] [-debug]\n\n";
  cerr << "\t This program will read in a .pts file specifying the \n";
  cerr << "\t the x/y/z/s components for each point separated by \n";
  cerr << "\t white space.  The input file should NOT contain any \n";
  cerr << "\t header.  The 's' component of each line is the string \n";
  cerr << "\t associated with that point. The SCIRun output file is \n";
  cerr << "\t written in ASCII unless -binOutput is specified.\n\n";
}

int
main(int argc, char **argv)
{
  if (argc < 3 || argc > 6) {
    printUsageInfo(argv[0]);
    return 2;
  }

  SCIRunInit();
  setDefaults();

  const char *ptsName = argv[1];
  const char *fieldName = argv[2];
  if (!parseArgs(argc, argv)) {
    printUsageInfo(argv[0]);
    return 2;
  }

  ProgressReporter *pr = scinew ProgressReporter();
  FieldHandle pcH(TextPointCloudString_reader(pr, ptsName));
  
  if (pcH.get_rep() == 0)
  {
    return 2;
  }

  if (binOutput) {
    BinaryPiostream out_stream(fieldName, Piostream::Write);
    Pio(out_stream, pcH);
  } else {
    TextPiostream out_stream(fieldName, Piostream::Write);
    Pio(out_stream, pcH);
  }
  return 0;  
}    
